/**
 *  Copyright (c) 2008-2011 Sonatype, Inc. All rights reserved.
 *
 *  This program is licensed to you under the Apache License Version 2.0,
 *  and you may not use this file except in compliance with the Apache License Version 2.0.
 *  You may obtain a copy of the Apache License Version 2.0 at http://www.apache.org/licenses/LICENSE-2.0.
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the Apache License Version 2.0 is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the Apache License Version 2.0 for the specific language governing permissions and limitations there under.
 */
package org.sonatype.jira.recorder;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.net.URL;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.apache.xmlrpc.XmlRpcException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.sonatype.jira.JiraXmlRpc;
import org.sonatype.jira.JiraXmlRpcProxy;

@SuppressWarnings( "rawtypes" )
public class Replayer
    implements InvocationHandler, JiraXmlRpcProxy
{

    private final Logger logger = LoggerFactory.getLogger( this.getClass() );

    private final JiraXmlRpc proxy;

    private URL url;

    private final List<Triple> triplets;

    public Replayer( final File log )
        throws IOException
    {
        triplets = Recorder.read( log );
        this.proxy = (JiraXmlRpc) Proxy.newProxyInstance( this.getClass().getClassLoader(), new Class[] { JiraXmlRpc.class }, this );
    }

    @Override
    public Object invoke( final Object proxy, final Method method, final Object[] args )
        throws Throwable
    {
        String methodName = method.getName();
        logger.trace( "replaying rpc method: {}#{}({})", new Object[] { url, methodName, Arrays.toString( args ) } );

        Object result = null;
        for ( Iterator<Triple> iterator = triplets.iterator(); iterator.hasNext(); )
        {
            Triple triple = iterator.next();
            if ( triple.getMethod().equals( methodName ) && Arrays.deepEquals( triple.getArgs(), args ) )
            {
                iterator.remove();
                result = triple.getResult();

                logger.trace( "returning recorded rpc result for method: {}#{}({})\n{}", new Object[] { url,
                    methodName, Arrays.toString( args ), result } );
                break;
            }
        }

        if ( result == null )
        {
            throw new IllegalStateException( "Could not find recorded answer" );
        }

        if ( result instanceof Object[] )
        {
            Vector<Object> returnValue = new Vector<Object>();
            Object[] resultArray = (Object[]) result;
            for ( Object o : resultArray )
            {
                if ( o instanceof Map )
                {
                    Hashtable<Object, Object> hash = new Hashtable<Object, Object>();
                    hash.putAll( (Map<?, ?>) o );
                    returnValue.add( hash );
                }
                else
                {
                    returnValue.add( o );
                }
            }
            result = returnValue;
        }
        else if ( result instanceof Map )
        {
            Hashtable<Object, Object> hash = new Hashtable<Object, Object>();
            hash.putAll( (Map<?, ?>) result );
            result = hash;
        }
        else if ( result instanceof XmlRpcException )
        {
            throw (XmlRpcException) result;
        }

        Class<?> returnType = method.getReturnType();

        return returnType.cast( result );
    }

    @Override
    public boolean addComment( final String token, final String issueKey, final String comment )
    {
        return proxy.addComment( token, issueKey, comment );
    }

    @Override
    public Hashtable createIssue( final String token, final Hashtable rIssueStruct )
    {
        return proxy.createIssue( token, rIssueStruct );
    }

    @Override
    public Vector getComments( final String token, final String issueKey )
    {
        return proxy.getComments( token, issueKey );
    }

    @Override
    public Vector getComponents( final String token, final String projectKey )
    {
        return proxy.getComponents( token, projectKey );
    }

    @Override
    public Vector getFavouriteFilters( final String token )
    {
        return proxy.getFavouriteFilters( token );
    }

    @Override
    public Hashtable getIssue( final String token, final String issueKey )
    {
        return proxy.getIssue( token, issueKey );
    }

    @Override
    public Vector getIssuesFromFilter( final String token, final String filterId )
    {
        return proxy.getIssuesFromFilter( token, filterId );
    }

    @Override
    public Vector getIssuesFromTextSearch( final String token, final String searchTerms )
    {
        return proxy.getIssuesFromTextSearch( token, searchTerms );
    }

    @Override
    public Vector getIssuesFromTextSearchWithProject( final String token, final Vector projectKeys, final String searchTerms,
                                                      final int maxNumResults )
    {
        return proxy.getIssuesFromTextSearchWithProject( token, projectKeys, searchTerms, maxNumResults );
    }

    @Override
    public Vector getIssueTypes( final String token )
    {
        return proxy.getIssueTypes( token );
    }

    @Override
    public Vector getIssueTypesForProject( final String token, final String projectId )
    {
        return proxy.getIssueTypesForProject( token, projectId );
    }

    @Override
    public Vector getPriorities( final String token )
    {
        return proxy.getPriorities( token );
    }

    @Override
    public Vector getProjectsNoSchemes( final String token )
    {
        return proxy.getProjectsNoSchemes( token );
    }

    @Override
    public Vector getResolutions( final String token )
    {
        return proxy.getResolutions( token );
    }

    @Override
    @Deprecated
    public Vector getSavedFilters( final String token )
    {
        return proxy.getSavedFilters( token );
    }

    @Override
    public Hashtable getServerInfo( final String token )
    {
        return proxy.getServerInfo( token );
    }

    @Override
    public Vector getStatuses( final String token )
    {
        return proxy.getStatuses( token );
    }

    @Override
    public Vector getSubTaskIssueTypes( final String token )
    {
        return proxy.getSubTaskIssueTypes( token );
    }

    @Override
    public Vector getSubTaskIssueTypesForProject( final String token, final String projectId )
    {
        return proxy.getSubTaskIssueTypesForProject( token, projectId );
    }

    @Override
    public Hashtable getUser( final String token, final String username )
    {
        return proxy.getUser( token, username );
    }

    @Override
    public Vector getVersions( final String token, final String projectKey )
    {
        return proxy.getVersions( token, projectKey );
    }

    @Override
    public String login( final String username, final String password )
    {
        return proxy.login( username, password );
    }

    @Override
    public Boolean logout( final String token )
    {
        return proxy.logout( token );
    }

    @Override
    public Hashtable updateIssue( final String token, final String issueKey, final Hashtable fieldValues )
    {
        return proxy.updateIssue( token, issueKey, fieldValues );
    }

    @Override
    public void setProxiedUrl( final String url )
    {
    }

    @Override
    public String getProxiedUrl()
    {
        return "http://replayer.invalid";
    }

}
